--[[
Name: Avion-1.0
Revision: $Revision: 1 $
Author(s): Sirow (EU-Gul'dan)
Documentation: http://www.wowace.com/wiki/Avion-1.0
SVN: http://svn.wowace.com/wowace/trunk/LibAvion/
Description: Allows to trigger visual effects (Library).
Dependencies: AceLibrary, AceEvent-2.0
]]


-------------------------------------------------------------------------------
-- Preamble
-------------------------------------------------------------------------------
-- Version control
local MAJOR_VERSION = "Avion-1.0"
local MINOR_VERSION = "$Revision: 2 $"

-- Take a loot at main-dependency
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

-- Make sure sup-dependencies are also there
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0.") end

-- Library holder
local AceEvent 	= AceLibrary("AceEvent-2.0")

local Avion = {}
AceEvent:embed(Avion)


-------------------------------------------------------------------------------
-- Library - API
-------------------------------------------------------------------------------
-- "Register" a new frame for animation (this registration handels settings)
function Avion:TriggerAnimation(triggerID, animationTable)
	-- Quite when disabled or allready registered
	if (not self.vars.running or self:IsAnimationRunning(triggerID)) then
		return
	end
	
	-- Make things local
	local object
	object, self.vars.animationCache[triggerID] = self:GetAvailableFrame()
	
	-- Save which triggerID triggerd this animation
	object.triggerID = triggerID
	
	-- Play soundfile if there is one to play
	if (animationTable.soundFile) then
		PlaySoundFile(animationTable.soundFile)
	end
	
	-- Finally "start" animation/timer
	self:StartAnimationTimer(object, animationTable, animationTable.animationTime, animationTable.waitTime)
end


-- "Unregister" animation
function Avion:StopAnimation(triggerID)
	-- Hmmm, well trying to unregister an unregister animation may be painful >:>
	if (self:IsAnimationRunning(triggerID)) then		
		-- Unregister animation (delayed and repeating animations)
		if (self:IsEventScheduled("Avion_UpdateAnimation_" .. triggerID)) then
			self:CancelScheduledEvent("Avion_UpdateAnimation_" .. triggerID)
		end
		if (self:IsEventScheduled("Avion_DisplayAnimation_" .. triggerID)) then
			self:CancelScheduledEvent("Avion_DisplayAnimation_" .. triggerID)
		end
		if (self:IsEventScheduled("Avion_WaitAnimation_" .. triggerID)) then
			self:CancelScheduledEvent("Avion_WaitAnimation_" .. triggerID)
		end
		
		-- Hide frame and its not inUse anymore...
		self.vars.frameCache[self.vars.animationCache[triggerID]].texture:SetTexture("Interface\\Icons\\Temp.blp")
		self.vars.frameCache[self.vars.animationCache[triggerID]].frame:Hide()
		self.vars.frameCache[self.vars.animationCache[triggerID]].inUse = false
		
		-- Delete frame from animation cache
		self.vars.animationCache[triggerID] = nil
	end
end



-- "Unregister" ALL animations
function Avion:StopAllAnimations()
	if (self.vars.animationCache) then
		for triggerID, _ in pairs(self.vars.animationCache) do
			self:StopAnimation(triggerID)
		end
	end
end


-- Take a look, if Animation isn't allready registered
function Avion:IsAnimationRunning(triggerID)
	if (self.vars.animationCache and self.vars.animationCache[triggerID]) then
		return true
	else
		return false
	end
end


-------------------------------------------------------------------------------
-- Library - Core
-------------------------------------------------------------------------------
-- "Register" a new frame for animation (this registration handels timing)
function Avion:StartAnimationTimer(object, animationTable, animationTime, waitTime)
	-- If there is NO wait-time, initialize and run animation
	if (waitTime == 0) then
		-- Initialize frame stuff
		if (not object.frame:IsVisible()) then
			self:InitFrame(object, animationTable)
		end
		
		-- We only need to test if something is animated if there is really animation-time ^^
		local anythingAnimated = false
		if (animationTime ~= 0) then
			-- If there is nothing to animate, well ... life is easier ^^
			for typ, value in pairs(animationTable.isAnimated) do
				if (value) then
					anythingAnimated = true
					break
				end
			end
		end
		
		-- Well ... thats life, go the hard way ^^
		if (animationTime == 0 or anythingAnimated) then
			-- Register animation
			self:ScheduleRepeatingEvent(
				"Avion_UpdateAnimation_" ..object.triggerID,
				self.UpdateAnimation, 
				self.vars.updateInterval, 
				self, 
				object, 
				animationTable
			)
		else
			-- Start animation after animationTime has passed, since there was nothing to animate...
			self:ScheduleEvent(
				"Avion_DisplayAnimation_" .. object.triggerID,
				self.StartAnimationTimer, 
				animationTable.animationTime,
				self, 
				object, 
				animationTable, 
				0, 
				0
			)
		end
	-- If there IS wait-time, retrigger after waitTime-delay
	else
		-- Show frame allready?
		if (animationTable.showWhileWait) then
			self:InitFrame(object, animationTable)
		end
		
		-- Start waiting
		self:ScheduleEvent(
			"Avion_WaitAnimation_" .. object.triggerID,
			self.StartAnimationTimer, 
			animationTable.waitTime, 
			self, 
			object, 
			animationTable, 
			animationTime, 
			0
		)
	end
end


-- Get an unused frame or create a new one
function Avion:GetAvailableFrame()
	local frameID = self:GetAvailableFrameID()
	return self.vars.frameCache[frameID], frameID
end


-- Get an unused frame or create a new one (only ID)
function Avion:GetAvailableFrameID()
	-- Is there any unused/invisible frame?
	for index, cache in pairs(self.vars.frameCache) do
		if (not cache.inUse) then
			-- Enable inUse and return index
			cache.inUse = true
			cache.elapsedTimes = 0
			
			return index
		end
	end
	
	-- If not, create one!
	return self:MakeFrame()
end


-- Create a new frame
function Avion:MakeFrame()
	-- Name of the frame
	local frameName = "Avion_Frame_" .. self.vars.frameCounter

	-- Create a new frame and apply some basic settings
	local frame = CreateFrame("FRAME", frameName, UIParent)
	frame:Hide()
	frame:SetPoint("CENTER", UIParent, "CENTER")
	frame:SetBackdropColor(0, 0, 0, 0)
	frame:SetFrameStrata("BACKGROUND")
	frame:SetScale(1)
	
	-- Add a texture to the frame (Blend mode = additive)
	local texture = frame:CreateTexture(frameName .. "_Texture", "BACKGROUND")
	texture:SetParent(frame)
	texture:SetAllPoints(frame)
	texture:SetBlendMode("ADD")
	
	-- Add everything to the cache
	self.vars.frameCache[self.vars.frameCounter] = {}
	self.vars.frameCache[self.vars.frameCounter].inUse = true
	self.vars.frameCache[self.vars.frameCounter].frame = frame
	self.vars.frameCache[self.vars.frameCounter].texture = texture
	self.vars.frameCache[self.vars.frameCounter].elapsedTimes = 0
	
	-- Increment number of frames and return new number
	self.vars.frameCounter = self.vars.frameCounter + 1
	return (self.vars.frameCounter-1)
end


-- Initialize frame settings and show it
function Avion:InitFrame(object, animationTable)
	-- (Re)set texture file and color
	object.texture:SetTexture(animationTable.textureFile)
	object.texture:SetVertexColor(animationTable.color.r, animationTable.color.g, animationTable.color.b)
	
	-- Adjust frame size, position and alpha
	object.frame:SetHeight(animationTable.size.height * 1.4142)
	object.frame:SetWidth(animationTable.size.width * 1.4142)
	object.frame:SetPoint("CENTER", UIParent, "CENTER", animationTable.position.x, animationTable.position.y)
	object.frame:SetAlpha(animationTable.alpha)
	
	-- Rotation
	local cosAlpha = 0.7072 * math.cos(animationTable.rotation)
	local sinAlpha = 0.7072 * math.sin(animationTable.rotation)
	object.texture:SetTexCoord(
		-cosAlpha + sinAlpha + 0.5, -sinAlpha - cosAlpha + 0.5, 
		-cosAlpha - sinAlpha + 0.5, -sinAlpha + cosAlpha + 0.5, 
		 cosAlpha + sinAlpha + 0.5,  sinAlpha - cosAlpha + 0.5, 
		 cosAlpha - sinAlpha + 0.5,  sinAlpha + cosAlpha + 0.5
	)
	
	-- Finally show it
	object.frame:Show()
end


-- Update frame animation
function Avion:UpdateAnimation(object, animationTable, forceUpdate)
	-- Update elapsed time (time since animation was triggered)
	local elapsedTime = object.elapsedTimes * self.vars.updateInterval
	object.elapsedTimes = object.elapsedTimes + 1
	
	-- Parameter for update function
	local param
	
	-- Animate frame, move, resize, color, rotate, ect.
	if (elapsedTime <= animationTable.animationTime) then
		-- Calc parameter for linear interpolation (param \in [0...1]) and animate frame
		if (animationTable.animateAndFade) then
			param = elapsedTime / (animationTable.animationTime + animationTable.fadeTime)
		else
			param = elapsedTime / animationTable.animationTime
		end
		self:AnimateFrame(object, animationTable, param)
		
	-- Start fading frame
	elseif (elapsedTime - animationTable.animationTime <= animationTable.fadeTime) then	
		-- Calc parameter for linear interpolation (param \in [0...1]) and continue animation
		if (animationTable.animateAndFade) then
			param = elapsedTime / (animationTable.animationTime + animationTable.fadeTime)
			
			self:AnimateFrame(object, animationTable, param)
		elseif (forceUpdate) then
			self:AnimateFrame(object, animationTable, 1)
		end

		-- Fade frame out
		param = (elapsedTime - animationTable.animationTime)/ animationTable.fadeTime
		object.frame:SetAlpha(
			self:Interpolation(
				animationTable.alpha, 
				0, 
				animationTable.animationFunc.alpha(param, animationTable.addRandom.alpha)
			)
		)
	
	-- Done, Unregister
	else
		self:StopAnimation(object.triggerID)
	end
end


-- Actually update frame animation
function Avion:AnimateFrame(object, animationTable, param)
	-- Get and set updated values
	if (animationTable.isAnimated.position) then
		local position = {
			x = self:Interpolation(
				animationTable.position.x, 
				animationTable.animation.position.x, 
				animationTable.animationFunc.position(param, animationTable.addRandom.position)
			),
			y = self:Interpolation(
				animationTable.position.y, 
				animationTable.animation.position.y, 
				animationTable.animationFunc.position(param, animationTable.addRandom.position)
			)
		}
		object.frame:SetPoint("CENTER", UIParent, "CENTER", position.x, position.y)
	end
	if (animationTable.isAnimated.scale) then
		local scale = self:Interpolation(
			1, 
			animationTable.animation.scale, 
			animationTable.animationFunc.scale(param, animationTable.addRandom.scale)
		)
		object.frame:SetHeight(animationTable.size.height * 1.4142 * scale)
		object.frame:SetWidth(animationTable.size.width * 1.4142 * scale)
	end
	if (animationTable.isAnimated.color) then
		local color = {
			r = self:Interpolation(
				animationTable.color.r, 
				animationTable.animation.color.r, 
				animationTable.animationFunc.color(param, animationTable.addRandom.color)
			),
			g = self:Interpolation(
				animationTable.color.g, 
				animationTable.animation.color.g, 
				animationTable.animationFunc.color(param, animationTable.addRandom.color)
			),
			b = self:Interpolation(
				animationTable.color.b, 
				animationTable.animation.color.b, 
				animationTable.animationFunc.color(param, animationTable.addRandom.color)
			)
		}
		object.texture:SetVertexColor(color.r, color.g, color.b)
	end
	if (animationTable.isAnimated.rotation) then
		local alpha = self:Interpolation(
			animationTable.rotation, 
			animationTable.animation.rotation,
			animationTable.animationFunc.rotation(param, animationTable.addRandom.rotation)
		)
		-- If someone wants to have this explained, you are free to ask, it's really easy ;)
		local cosAlpha = 0.7072 * math.cos(alpha)
		local sinAlpha = 0.7072 * math.sin(alpha)
		
		object.texture:SetTexCoord(
			-cosAlpha + sinAlpha + 0.5, -sinAlpha - cosAlpha + 0.5, 
			-cosAlpha - sinAlpha + 0.5, -sinAlpha + cosAlpha + 0.5, 
			 cosAlpha + sinAlpha + 0.5,  sinAlpha - cosAlpha + 0.5, 
			 cosAlpha - sinAlpha + 0.5,  sinAlpha + cosAlpha + 0.5
		)
	end	
end


-- Update values with linear interporlation between start and end
function Avion:Interpolation(startValues, endValues, param) 
	return startValues + (endValues - startValues) * param
end


-------------------------------------------------------------------------------
-- Activation function und embedding
-------------------------------------------------------------------------------
-- Activate a new instance of this library
local function activate(self, oldLib, oldDeactivate)
	-- If there is an old library running, disable and copy everything
	if oldLib then
		-- Copy vars from old library
		self.vars = oldLib.vars
		
		-- Disable old librarys event listender
		oldLib:StopAllAnimations()
	else
		self.vars = {}
		self.vars.animationCache = {}
		self.vars.updateInterval = 0.05
		self.vars.frameCache = {}
		self.vars.frameCounter = 1
	end

	-- (Re-) initialize variables
	self.vars.running = true

	if oldDeactivate then oldDeactivate(oldLib) end
end


-------------------------------------------------------------------------------
-- Register
-------------------------------------------------------------------------------
AceLibrary:Register(Avion, MAJOR_VERSION, MINOR_VERSION, activate)
Avion = nil
